#include "merlin.h"


#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "LUFA/Drivers/USB/USB.h"
#include "Keyboard.h"


#include "Q/osc/osc.h"


#ifndef F_USB
#include <avr/iox128a1u.h>
#endif


merlin_s merlin =
{
	false,
	true,
	{},
	0,
	{0,0,0,0},
	0,
	0,
	0,
	false,
	false, //Mouse mode
	1,
	0,
	0,
	0,
	true
};

merlin_mouse_s mouse = 
{
	80,
	79,
	82,
	81,
	88,
	92,
};

//TODO: Timings are not dynamicaly calced


int main(void)
{

	//################################################################################################
	// Initialisieren aller Komponenten und Variablen
	//################################################################################################
	PMIC.CTRL = PMIC_LOLVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm;
	int2x_ppl_USB48(18);

	sei();

	PORTF.DIR = 0xFF;

	merlin_capture_init();
	
	USB_Init();

	CDC_Device_CreateStream(&VirtualSerial_CDC_Interface, stdout);
	
	while(1)
	{
		CDC_Device_ReceiveByte(&VirtualSerial_CDC_Interface);
		
		CDC_Device_USBTask(&VirtualSerial_CDC_Interface);
		HID_Device_USBTask(&Mouse_HID_Interface);
		HID_Device_USBTask(&Keyboard_HID_Interface);
		USB_USBTask();
	}

	return 0;
};








void merlin_capture_init(void)
{
	TCD0.PER = ~0;
	TCD0.CTRLB = TC_WGMODE_NORMAL_gc;
	
	TCD0.CTRLA = TC_CLKSEL_DIV1_gc;
	TCD0.INTCTRLA = TC_OVFINTLVL_MED_gc;
	
	PORTD.INT0MASK = PIN0_bm;
	
	PORTD.INTCTRL = PORT_INT0LVL_HI_gc;
	
	PORTD.PIN0CTRL = PORT_ISC_BOTHEDGES_gc;
}

// void merlin_translate(uint8_t *keys, uint8_t n_keys, uint8_t *mods)

void merlin_eval(void)
{
	uint8_t i, j;
	uint32_t time = 0;
	uint16_t p;
	uint8_t not_level = 1;
	uint16_t header = 0;
	uint8_t odd = 0;
	uint8_t count = 0;
	uint8_t use_data_bit_count = 0;
	uint8_t use_data_byte_count = 0;
	//Max 4 codes + toggle bit
	uint8_t codes[5];
	uint8_t n_keys = 0;
	uint8_t mod_byte = 0;
	uint8_t tmp;
	
	//Preamble 10000|00001
	for(i = 0; i < merlin.data_pos - 1; i++)
	{
		//Das Singal wird als Manchestercode betrachtet. Das bedeutet Folgendes:
		//Alle ungeraden Positionen sind Flanken.
		//Eine steigende Flanke -> 1
		//Eine fallende Flanke  -> 0
		
		p = (uint16_t)(merlin.data[i + 1] - merlin.data[i]);

		//################################### Fehlerfall ######################################
		if((!odd && p >= MERLIN_MAX_SHORT_LEVEL) || (p >= MERLIN_MAX_LEVEL))
		{
			//Fehlerhaft, wenn das Signal an einer geraden Position fuer 2 Pulslängen auf dem gleichen Level bleibt -> Keine Flanke an der ungeraden Position -> Es kann kein Bit gelesen werden.
			//Oder Fehlerhaft, weil das Signal fuer ungefaehr 3 oder mehr Pulslängen ohne Flanke ist. Das ist im Manchestercode unmöglich.
			
			//Da einige Infrarot-Sensoren eine Weile brauchen, um sich zu kalibrieren, akzeptieren wir es , falls die ersten Bits des Headers defekt sind.
			//Wenn weniger als 3 Bits gelesen wurden, ueberspringe die ersten 7 Bits, also 13,5 Perioden und mache bei der nächsten Flanke weiter.
			if(count >= 3)
			{
				return;
			}
			

			uint32_t time2 = 0;
			bool found = false;

			not_level = 1;
			//Wir ueberspringen die ersten 14 Pulse und lesen von dort weiter. So hatte der Sensor Zeit, sich zu kalibrieren.
			for(j = 0; j < merlin.data_pos - 1; j++)
			{
				time2 += (uint16_t)(merlin.data[j + 1] - merlin.data[j]);
				not_level = ~not_level & 0b1;
				if(time2 > MERLIN_SKIP_WIDTH)
				{
					header = 0b1000000;
					count = 7;
					found = true;
					break;
				}
			}
			
			if(found)
			{
				i = j;
				odd = 0;
				if(not_level)
				{
					return;
				}
				continue;
			}
			//Header konnte nicht gefunden werden
			return;

		}
		//#####################################################################################


		//Wenn die Zeitperiode < als 1,5 Perioden ist, gehen wir von *einer* Pulsbreite aus, und wir muessen "ungerade" (odd) togglen.
		if(p < MERLIN_MAX_SHORT_LEVEL)
		{
			odd = !odd;
		}
		
		//Wenn wir ungerade sind, muss das Bit gespeichert werden.
		//Dabei interessiert uns das aktuelle Signallevel, weil dadurch klar wird, ob es sich
		//um eine steigende oder fallende Flanke handelt.
		if(odd)
		{
			if(count < 10) //Die ersten 10 Bits werden im Header gespeichert
			{
				header <<= 1;
				header |= not_level;
			}
			else //Den Rest legen wir byte-weise ab.
			{    //Jedes Byte wird ein Keycode oder Modifier beinhalten
				codes[use_data_byte_count] <<= 1;
				codes[use_data_byte_count] |= not_level;
				use_data_bit_count++;
				if(use_data_bit_count == 8)
				{ //Bei Bit 8 springen wir zum nächsten Byte
					use_data_bit_count = 0;
					use_data_byte_count++;
				}
			}
			
			count++;
		}
		
		not_level = ~not_level & 0b1;

		//Wird nicht genutzt, könnte aber in Zukunft für eine Kalibrierung benutzt werden.
		time += p;
	}

	//Ist der Header fehlerhaft, ignorieren wir die Daten.
	if(header != 0b1000000001)
	{
		return;
	}
	
	//Das letzte reinkommenden Bit ist invers zum Vorletzten, ansonsten hat es keine Bedeutung.
	//Ist das nicht der Fall, gehen wir von Fehlern aus.
	//Die Stopmaske besteht allerdings nur aus dem Header ohne Togglebit
	if(use_data_byte_count > 0 && codes[use_data_byte_count] == (codes[use_data_byte_count - 1] & 0b1))
	{
		return;
	}
	
	bool mouse_move = false;
	uint8_t mod_key;
	uint8_t mouse_button = 0;
	
	//Die Keycodes werden intepretiert:
	for(i = 0; i < use_data_byte_count; i++)
	{
		tmp = codes[i];
		printf("%d\n\r", tmp);
		if(!tmp)
		{
			return;
		}
		
		//Das erste Bit gibt an, ob es sich um einen Modifier oder um einen Keycode handelt.
		mod_key = tmp & 0b10000000;
		tmp &= 0b01111111;
		
		
		if(mod_key)
		{  //Modifier
			switch(tmp)
			{
				case 96:
					mod_byte |= HID_KEYBOARD_MODIFIER_LEFTCTRL;
					break;
				case 97:
					mod_byte |= HID_KEYBOARD_MODIFIER_LEFTSHIFT;
					break;
				case 98:
					mod_byte |= HID_KEYBOARD_MODIFIER_LEFTALT;
					break;
				case 102:
					mod_byte |= HID_KEYBOARD_MODIFIER_RIGHTALT;
					break;
				default:
					return;
			}
		}
		else
		{ //Keycode
			
			//Erst die maus pruefen da sie nicht mit der keycode tabelle arbeitet
			if(merlin.mouse_mode)
			{
				if(tmp == mouse.left_key)
				{
					merlin.mouse_move_x = -merlin.mouse_jump;
					mouse_move = true;
					continue;
				}
				if(tmp == mouse.right_key)
				{
					merlin.mouse_move_x = merlin.mouse_jump;
					mouse_move = true;
					continue;
				}
				if(tmp == mouse.down_key)
				{
					merlin.mouse_move_y = merlin.mouse_jump;
					mouse_move = true;
					continue;
				}
				if(tmp == mouse.up_key)
				{
					merlin.mouse_move_y = -merlin.mouse_jump;
					mouse_move = true;
					continue;
				}
				if(tmp == mouse.b1)
				{
					mouse_button |= (1 << 0);
					continue;
				}
				if(tmp == mouse.b2)
				{
					mouse_button |= (1 << 1);
					continue;
				}
				//TODO: Dann kümmer dich um den hid report
			}
			
			//dann die keycodes
			codes[n_keys] = pgm_read_byte(merlin_keycodes + (tmp & 0b01111111));
			switch(codes[n_keys])
			{
				case 0:
					return;
				case 1:
					//Spezial key zum aktivieren/deaktivieen der Maus
					if(merlin.toggle_ready)
					{
						merlin.mouse_mode = !merlin.mouse_mode;
						printf_P(PSTR("m: %d\n\r"), merlin.mouse_mode);
						merlin.toggle_ready = false;
					}
					continue;
				default:
					n_keys++;
					continue;
			}
			

		}
	}
	
	//Vielleicht gibt es keine Keycodes, dann führen wir einen Stop durch.
	if(use_data_bit_count == 0 && use_data_byte_count == 0)
	{
		merlin.write_in_progres = true;
		merlin.n_keys = 0;
		merlin.mod_byte = 0;
		merlin.write_in_progres = false;
		merlin.mouse_jump = 1;
		merlin.buttons_pressed = 0;
		return;
	}
	
	//Anderenfalls tragen wir die Ergebnisse zur spaeteren Uebertragung in die globale Struktur ein.
	//Siehe Keyboard.c
	merlin.write_in_progres = true;
	for(i = 0; i < n_keys; i++)
	{
		merlin.keys[i] = codes[i];
	}
	merlin.n_keys = n_keys;
	merlin.mod_byte = mod_byte;
	if(mouse_move)
	{
		if(merlin.mouse_jump < 40)
		{
			merlin.mouse_jump++;
		}
	}
	else
	{
		merlin.mouse_jump = 1;
	}
	
	merlin.buttons_pressed = mouse_button;
	merlin.write_in_progres = false;
}




ISR(TCD0_OVF_vect)
{
	//Wenn eine Messung angefangen hat, aber kein Signal ueber einen gesammten
	//Timer-Zyklus gekommen ist, werten wir die Daten aus.
	if(!merlin.signal && merlin.log)
	{
		PORTD.INTCTRL = PORT_INT0LVL_OFF_gc;
		merlin_eval();
		merlin.log = false;
		merlin.data_pos = 0;
		merlin.timeout = 0;
		PORTD.INTCTRL = PORT_INT0LVL_HI_gc;
	}
	else
	{
		//Timeout-Handling, damit falls das Stopsignal nicht empfangen wird, der Empfaenger nicht immer
		//weiter Daten an den Host sendet.
		merlin.timeout++;
		if(merlin.timeout >= MERLIN_TIMEOUT)
		{
			merlin.write_in_progres = true;
			
			merlin.n_keys = 0;
			merlin.mod_byte = 0;
			merlin.timeout = 0;
			merlin.mouse_jump = 1;
			merlin.buttons_pressed = 0;
			merlin.toggle_ready = true;
			
			merlin.write_in_progres = false;
		}
	}
	
	//Setze "signal = false". Wird vom Toggle-Interrupt wieder auf "true" gesetzt.
	merlin.signal = false;
}

ISR(PORTD_INT0_vect)
{
	//Wir schreiben alle Flanken auf; aber keine Level. 
	//Diese lassen sich aus dem Wissen ableiten, dass der Startzustand "high" ist.
	uint16_t res = TCD0.CNT;
	if(merlin.data_pos < MERLIN_MAX_DATA_LENGTH)
	{
		merlin.data[merlin.data_pos] = res;
		merlin.data_pos++;
	}
	else
	{
		printf("warning no mem\n\r");
	}
	merlin.signal = true;
	merlin.log = true;
}


//Keycode Map
const uint8_t merlin_keycodes[128] PROGMEM =
{
	0, /*0*/
	0, /*1*/
	0, /*2*/
	0, /*3*/
	HID_KEYBOARD_SC_A, /*4*/
	HID_KEYBOARD_SC_B, /*5*/
	HID_KEYBOARD_SC_C, /*6*/
	HID_KEYBOARD_SC_D, /*7*/
	HID_KEYBOARD_SC_E, /*8*/
	HID_KEYBOARD_SC_F, /*9*/
	HID_KEYBOARD_SC_G, /*10*/
	HID_KEYBOARD_SC_H, /*11*/
	HID_KEYBOARD_SC_I, /*12*/
	HID_KEYBOARD_SC_J, /*13*/
	HID_KEYBOARD_SC_K, /*14*/
	HID_KEYBOARD_SC_L, /*15*/
	HID_KEYBOARD_SC_M, /*16*/
	HID_KEYBOARD_SC_N, /*17*/
	HID_KEYBOARD_SC_O, /*18*/
	HID_KEYBOARD_SC_P, /*19*/
	HID_KEYBOARD_SC_Q, /*20*/
	HID_KEYBOARD_SC_R, /*21*/
	HID_KEYBOARD_SC_S, /*22*/
	HID_KEYBOARD_SC_T, /*23*/
	HID_KEYBOARD_SC_U, /*24*/
	HID_KEYBOARD_SC_V, /*25*/
	HID_KEYBOARD_SC_W, /*26*/
	HID_KEYBOARD_SC_X, /*27*/
	HID_KEYBOARD_SC_Y, /*28*/
	HID_KEYBOARD_SC_Z, /*29*/
	HID_KEYBOARD_SC_1_AND_EXCLAMATION, /*30*/
	HID_KEYBOARD_SC_2_AND_AT, /*31*/
	HID_KEYBOARD_SC_3_AND_HASHMARK, /*32*/
	HID_KEYBOARD_SC_4_AND_DOLLAR, /*33*/
	HID_KEYBOARD_SC_5_AND_PERCENTAGE, /*34*/
	HID_KEYBOARD_SC_6_AND_CARET, /*35*/
	HID_KEYBOARD_SC_7_AND_AMPERSAND, /*36*/
	HID_KEYBOARD_SC_8_AND_ASTERISK, /*37*/
	HID_KEYBOARD_SC_9_AND_OPENING_PARENTHESIS, /*38*/
	HID_KEYBOARD_SC_0_AND_CLOSING_PARENTHESIS, /*39*/
	HID_KEYBOARD_SC_ENTER, /*40*/
	HID_KEYBOARD_SC_POWER, /*41*/
	HID_KEYBOARD_SC_BACKSPACE, /*42*/
	HID_KEYBOARD_SC_TAB, /*43*/
	HID_KEYBOARD_SC_SPACE, /*44*/
	0, /*45*/
	HID_KEYBOARD_SC_EQUAL_AND_PLUS, /*46*/
	0, /*47*/
	0, /*48*/
	0, /*49*/
	0, /*50*/
	HID_KEYBOARD_SC_SEMICOLON_AND_COLON, /*51*/
	HID_KEYBOARD_SC_APOSTROPHE_AND_QUOTE, /*52*/
	0, /*53*/
	HID_KEYBOARD_SC_COMMA_AND_LESS_THAN_SIGN, /*54*/
	HID_KEYBOARD_SC_DOT_AND_GREATER_THAN_SIGN, /*55*/
	HID_KEYBOARD_SC_SLASH_AND_QUESTION_MARK, /*56*/
	0, /*57*/
	0, /*58*/
	HID_KEYBOARD_SC_MEDIA_CALCULATOR, /*59*/
	0, /*60*/
	HID_KEYBOARD_SC_MEDIA_BACKWARD, /*61*/
	HID_KEYBOARD_SC_MEDIA_PLAY, /*62*/
	HID_KEYBOARD_SC_MEDIA_FORWARD, /*63*/
	HID_KEYBOARD_SC_MEDIA_STOP, /*64*/
	HID_KEYBOARD_SC_VOLUME_UP, /*65*/
	HID_KEYBOARD_SC_VOLUME_DOWN, /*66*/
	0, /*67*/
	0, /*68*/
	0, /*69*/
	0, /*70*/
	0, /*71*/
	0, /*72*/
	0, /*73*/
	0, /*74*/
	0, /*75*/
	0, /*76*/
	0, /*77*/
	0, /*78*/
	HID_KEYBOARD_SC_RIGHT_ARROW, /*79*/
	HID_KEYBOARD_SC_LEFT_ARROW, /*80*/
	HID_KEYBOARD_SC_DOWN_ARROW, /*81*/
	HID_KEYBOARD_SC_UP_ARROW, /*82*/
	0, /*83*/
	MOUSE_SWITCH_KEYCODE, /*84*/
	HID_KEYBOARD_SC_ESCAPE, /*85*/
	HID_KEYBOARD_SC_HELP, /*86*/
	0, /*87*/
	HID_KEYBOARD_SC_SELECT, /*88*/
	0, /*89*/
	0, /*90*/
	HID_KEYBOARD_SC_PAGE_DOWN, /*91*/
	HID_KEYBOARD_SC_PAGE_UP, /*92*/
	HID_KEYBOARD_SC_MEDIA_MUTE, /*93*/
	0, /*94*/
	HID_KEYBOARD_SC_NON_US_BACKSLASH_AND_PIPE, /*95*/
	HID_KEYBOARD_SC_MINUS_AND_UNDERSCORE, /*96*/ //HID_KEYBOARD_MODIFIER_LEFTCTRL
	HID_KEYBOARD_SC_F2, /*97*/ //HID_KEYBOARD_MODIFIER_LEFTSHIFT
	HID_KEYBOARD_SC_F5, /*98*/ //HID_KEYBOARD_MODIFIER_LEFTALT
	0, /*99*/
	HID_KEYBOARD_SC_MEDIA_EJECT, /*100*/
	0, /*101*/
	0, /*102*/ //HID_KEYBOARD_MODIFIER_RIGHTALT
	0, /*103*/
	0, /*104*/
	0, /*105*/
	0, /*106*/
	0, /*107*/
	0, /*108*/
	0, /*109*/
	0, /*110*/
	0, /*111*/
	0, /*112*/
	0, /*113*/
	0, /*114*/
	0, /*115*/
	0, /*116*/
	0, /*117*/
	0, /*118*/
	0, /*119*/
	0, /*120*/
	0, /*121*/
	0, /*122*/
	0, /*123*/
	0, /*124*/
	0, /*125*/
	0, /*126*/
	0, /*127*/
};